;; Generated package description from plz.el  -*- no-byte-compile: t -*-
(define-package "plz" "0.10pre0.20240816.164727" "HTTP library" '((emacs "27.1")) :commit "9fb22d84d9e39692392276d7b7b9979ca6ca1c01" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("comm" "network" "http") :url "https://github.com/alphapapa/plz.el")
